from dataclasses import dataclass, field


@dataclass(frozen=True, kw_only=True)
class Params:

    GC_weight: float = 7.0
    GT_weight: float = 1.0
    AC_weight: float = 1.0
    AT_weight: float = 0.5
    display_sequence_score: int = 0
    mismatch_penalty_starting_value: int = 1.0 
    mismatch_penalty_linear_delta: int = 2.0
    mismatch_penalty_type: str = "linear"
    mismatch_penalty_choices: tuple[str] = ("linear", "exponential")
    method_choices: tuple[str] = ("transitions", "coverage", "layered")
    consecutive_AT_scoring: tuple[int] = (0.5, 0.5, 0.5, 0.5, 0.0, 0.0, -5.0, -100.0)
    n_jobs: int = 8
    cadence_reward: float = 1
    method: str = "transitions"
    threshold: int = 50
    headers: list[str] = field(repr=False,
                               default_factory=lambda : ["Chromosome",
                                                         "Start",
                                                         "End",
                                                         "Z-DNA Score",
                                                         "Sequence",
                                                         "Total Sequence Score"])

    @property
    def __new_dict__(self) -> dict[str, int]:
        representation = self.__dict__.copy()
        representation.pop("headers")
        return representation



import uuid
from Bio import SeqIO
from pathlib import Path
import pandas as pd
import concurrent.futures
import os
from termcolor import colored
import time
from typing import Callable, Optional
import numpy as np
import functools
from .zdna_calculator import ZDNACalculatorSeq, Params
from collections.abc import Iterable
import logging
from attrs import define, field
import argparse


@define(kw_only=True, slots=True, frozen=True)
class ZDNASubmissionForm:
    recordSeq: str = field(converter=str)
    recordID: str = field(converter=str)
    input_fasta: os.PathLike[str] = field(converter=str)

def read_fasta(path: Path | str):
    for record in SeqIO.parse(path, "fasta"):
        yield (str(record.id), str(record.seq))

def timeit(func: Callable) -> Callable:
    @functools.wraps(func)
    def wrapper(*args, **kwargs):
        print(colored("ZDNA extraction initialized.", "magenta"))
        then = time.perf_counter()
        result = func(*args, **kwargs)
        now = time.perf_counter()
        print(colored(f"Process finished within {now-then:.2f} second(s).", "magenta"))
        return result
    return wrapper

def _extract(ID: int, sequence: str, params: Params) -> pd.DataFrame:
    zdna_string = ZDNACalculatorSeq(data=sequence, params=params)
    subarrays_detected = zdna_string.subarrays_above_threshold()
    subarrays_detected = pd.DataFrame(subarrays_detected, columns=params.headers[1:-1])
    scoring_array = zdna_string.scoring_array

    if not len(subarrays_detected):
        null_subarrays = pd.DataFrame([[np.nan]*len(params.headers)], columns=params.headers)
        # Explicitly set the 'Chromosome' column to object dtype to avoid FutureWarning
        null_subarrays['Chromosome'] = null_subarrays['Chromosome'].astype(object)
        null_subarrays.loc[:, "Chromosome"] = ID
        null_subarrays.loc[:, "Total Sequence Score"] = sum(scoring_array)
        return null_subarrays

    subarrays_detected.loc[:, "Chromosome"] = ID
    subarrays_detected.loc[:, "Total Sequence Score"] = sum(scoring_array)
    return subarrays_detected[params.headers]

def _subextraction(submission_forms: list[ZDNASubmissionForm], params: Params) -> pd.DataFrame:
    # process concurrently
    outputs = []
    for form in submission_forms:
        try:
            present_df = _extract(
                ID=form.recordID,
                sequence=form.recordSeq,
                params=params
            )
            outputs.append(present_df)
        except IndexError as e:
            logging.error(
                f"Failed to process input fasta '{Path(form.input_fasta).name}' on chromosome {str(form.recordID)}. "
                f"Chromosome length: {len(str(form.recordSeq))}."
            )
            logging.error(f"Error message: '{e}'.")
            continue

    if outputs:
        zdna_df = pd.concat(outputs, axis=0).reset_index(drop=True)
    else:
        zdna_df = pd.DataFrame(columns=params.headers)
        logging.info("Empty dataframe derived from subprocess _subextraction")
    return zdna_df

def assign_tasks(tasks: list, total_buckets: int) -> list[list]:
    then = time.perf_counter()
    total = len(tasks)
    step = total // total_buckets
    remainder = total % total_buckets
    assigned_tasks = []
    infimum = 0
    while True:
        if remainder > 0:
            supremum = infimum + step + 1
            remainder -= 1
        else:
            supremum = infimum + step
        assigned_tasks.append(tasks[infimum: supremum])

        if len(assigned_tasks) == total_buckets:
            break

        infimum = supremum

    now = time.perf_counter()

    # only for testing purposes
    try:
        assigned_numpy_tasks = [job.tolist() for job in np.array_split(tasks, total_buckets)]
    except Exception as e:
        logging.error(e)
        logging.error(f"Woops! Failed to use numpy splitting due to the following error '{e}'.")
        assigned_numpy_tasks = None

    if assigned_numpy_tasks is not None:
        assert assigned_tasks == assigned_numpy_tasks, "Invalid task assignment."

    print(colored(f"Task assignment completed within {now-then:.2f} second(s).", "green"))
    return assigned_tasks

def _postmerge_callback(future) -> None:
    # This function can be implemented if needed
    pass

def sniff(iterable: Iterable) -> tuple[tuple[str, str], Iterable]:
    try:
        first = next(iterable)
    except StopIteration:
        return None
    return first, itertools.chain([first], iterable)

def extract_zdna_v2( path: str | os.PathLike[str], params: Params, n_jobs: int = 1 ) -> pd.DataFrame:
    then = time.perf_counter()
    zdna_df = pd.DataFrame(columns=params.headers)
    outputs: list[pd.DataFrame] = []
    assert isinstance(n_jobs, int) and n_jobs > 0, "Number of jobs must be a positive int."

    submission_forms: list[ZDNASubmissionForm] = []
    for rec_id, rec_seq in read_fasta(path):
        submission_forms.append(
            ZDNASubmissionForm(
                recordSeq=rec_seq.upper(),
                recordID=rec_id,
                input_fasta=path
            )
        )
    

    # split into records
    assigned_tasks = assign_tasks(submission_forms, n_jobs)
    with concurrent.futures.ProcessPoolExecutor(max_workers=n_jobs) as executor:
        results = executor.map(_subextraction, assigned_tasks, [params]*n_jobs)

        for result in results:
            print("Task finished.")
            outputs.append(result)

    zdna_df = pd.concat(outputs, axis=0).reset_index(drop=True)

    now = time.perf_counter()
    print(f"Process finished within {now-then:.2f} second(s).")

    if params.display_sequence_score == 0:
        zdna_df.drop(columns=['Total Sequence Score'], inplace=True)
        zdna_df.dropna(inplace=True, axis=0)
    return zdna_df

@timeit
def transform( path: Path | str, params: Params, max_resources_threshold: int, n_jobs: int = 1 ) -> pd.DataFrame:
    logging.info(f"Processing file '{Path(path).name}'...")
    # lazy loading
    zdna_df = extract_zdna_v2(
        path,
        params,
        n_jobs=n_jobs
    )
    print(colored("Saving Z-DNA dataframe...", "magenta"))
    output_dir = Path("extractions_zdna_human")
    output_dir.mkdir(exist_ok=True)
    output_file = output_dir / f"{Path(path).stem}_zdna_score.csv"
    zdna_df.to_csv(
        output_file,
        mode="w",
        index=False
    )

    logging.info(f"File '{Path(path).name}' has been processed successfully.")
    return zdna_df

def main():

    parser = argparse.ArgumentParser(description="""Given a fasta file and the corresponding parameters
                                                    it calculates the ZDNA for each
                                                    sequence present.""")
    parser.add_argument("--path", type=str, default="test_file.fna", help="Path to file analyzed")
    parser.add_argument("--GC_weight", type=float, default=Params.GC_weight,
                        help=f"Weight given to GC and CG transitions. Default = {Params.GC_weight}")
    parser.add_argument("--AT_weight", type=float, default=Params.AT_weight,
                        help=f"Weight given to AT and TA transitions. Default = {Params.AT_weight}")
    parser.add_argument("--GT_weight", type=float, default=Params.GT_weight,
                        help=f"Weight given to GT and TG transitions. Default = {Params.GT_weight}")
    parser.add_argument("--AC_weight", type=float, default=Params.AC_weight,
                        help=f"Weight given to AC and CA transitions. Default = {Params.AC_weight}")
    parser.add_argument("--mismatch_penalty_starting_value", type=int, default=Params.mismatch_penalty_starting_value,
                        help=f"Penalty applied to the first non purine/pyrimidine transition encountered. "
                             f"Default = {Params.mismatch_penalty_starting_value}")
    parser.add_argument("--mismatch_penalty_linear_delta", type=int, default=Params.mismatch_penalty_linear_delta,
                        help=f"Only applies if penalty type is set to linear. Determines the rate of increase of the "
                             f"penalty for every subsequent non purine/pyrimidine transition. "
                             f"Default = {Params.mismatch_penalty_linear_delta}")
    parser.add_argument("--mismatch_penalty_type", choices=Params.mismatch_penalty_choices,
                        default=Params.mismatch_penalty_type,
                        help=f"Method of scaling the penalty for contiguous non purine/pyrimidine transition. "
                             f"Default = {Params.mismatch_penalty_type}")
    parser.add_argument("--method", choices=Params.method_choices, default=Params.method,
                        help=f"Method used for the Z-DNA scoring algorithm. See documentation for more details. "
                             f"Default = {Params.method}")
    parser.add_argument("--n_jobs", type=int, default=Params.n_jobs,
                        help="Number of threads to use. Defaults to -1, which uses the maximum available threads on CPU")
    parser.add_argument("--threshold", type=int, default=Params.threshold,
                        help=f"Scoring threshold for a for a sequence to be considered potentially Z-DNA forming and "
                             f"returned by the program. This parameter is also used for determining how big the scoring "
                             f"drop within a sequence should be, before it is split into two separate Z-DNA candidate "
                             f"sequences. Default={Params.threshold}")
    parser.add_argument("--consecutive_AT_scoring", type=Iterable[str], default=Params.consecutive_AT_scoring,
                        help=f"Consecutive AT repeats form a hairpin structure instead of Z-DNA. In order to reflect that, "
                             f"a penalty array is defined, which provides the score adjustment for the first and the "
                             f"subsequent TA appearances. The last element will be applied to every subsequent TA "
                             f"appearance. For more information see documentation. Default = {Params.consecutive_AT_scoring}")
    parser.add_argument("--max_resources_threshold", type=int, default=100_000)
    parser.add_argument("--display_sequence_score", type=int, choices=[0, 1], default=0)

    args = parser.parse_args()
    _params = vars(args)
    
    max_resources_threshold = _params.pop("max_resources_threshold")
    path = _params.pop("path")
    n_jobs = _params.get("n_jobs")

    for key in _params:
        if key in ("consecutive_AT_scoring", "method", "mismatch_penalty_type", "path"):
            continue
        assert _params[key] >= 0.0, "Params must be a non-negative integer."

    params = Params(**_params)
    assert Path(path).expanduser().resolve().is_file(), f"No file {path} was found."

    print(colored("Process parameters", "magenta"))
    print(colored("*" * 25, "magenta"))
    for key, value in params.__new_dict__.items():
        print(colored(f"{key}: {value}", "magenta"))

    print(colored("*" * 25, "magenta"))

    logging.basicConfig(
        level=logging.INFO,
        filemode="a",
        filename="zdna_extractions.log",
        format="%(levelname)s:%(message)s"
    )
    transform(path, params=params, n_jobs=n_jobs, max_resources_threshold=max_resources_threshold)

if __name__ == "__main__":
    main()


from Bio.Seq import Seq
import numpy as np

from typing import Optional
from .constants import Params

class ZDNACalculatorSeq(Seq):

    def __init__(self, data,  params: Optional[Params] = None) -> None:
        super().__init__(data.upper())
        self.scoring_array = []  # Calculate scoring array in subarrays method do do only when necessary
        if params is None:
            self.params = Params()
        else:
            self.params = params


    def zdna_calculator_layered(self):
        """
        Calculates the scoring array for a given sequence, using seperate scorring for transitions, cadence and TA penalty
        :return: np.ndarray: Scoring array
        """
        scoring_array = np.empty(len(self) - 1, dtype=float)
        mismatches_counter = 0  # k is the number of mismatches
        cadence = 1  # indicating if the alternative purine-pyrimidine pattern is unbroken
        consecutive_AT_counter = 0  # Number of consecutive TAs or ATs
        for i in range(len(self) - 1):
            match self[i] + self[i + 1]:
                case "GC" | "CG":
                    scoring_array[i] = self.params.GC_weight
                    mismatches_counter = 0  # Reset the number of mismatches
                case "GT" | "TG":
                    scoring_array[i] = self.params.GT_weight
                    mismatches_counter = 0
                case "AC" | "CA":
                    scoring_array[i] = self.params.AC_weight
                    mismatches_counter = 0
                case "AT" | "TA":
                    adjusted_weight = self.params.AT_weight
                    # If we have less consecutive ATs than the provided AT scoring list, use the corresponding element
                    # Else use the last element of the list as penalty
                    if consecutive_AT_counter < len(self.params.consecutive_AT_scoring):
                        adjusted_weight += self.params.consecutive_AT_scoring[consecutive_AT_counter]
                    else:
                        adjusted_weight += self.params.consecutive_AT_scoring[-1]
                    scoring_array[i] = adjusted_weight
                    consecutive_AT_counter += 1
                    mismatches_counter = 0
                case _:  # If there was a mismatch
                    mismatches_counter += 1
                    if self.params.mismatch_penalty_type == "exponential":
                        # Limit the maximum value of the mismatch penalty to avoid exploding values
                        scoring_array[i] = -self.params.mismatch_penalty_starting_value ** mismatches_counter if mismatches_counter < 15 else -32000
                    elif self.params.mismatch_penalty_type == "linear":
                        scoring_array[i] = -self.params.mismatch_penalty_starting_value - self.params.mismatch_penalty_linear_delta * (mismatches_counter-1)
                    else:
                        raise ValueError(f"Mismatch penalty type not recognized. Valid options are {Params.mismatch_penalty_choices}")
            if self[i] + self[i + 1] in ("GC", "CG", "GT", "TG", "AC", "CA", "AT", "TA"):
                scoring_array[i] += self.params.cadence_reward
        return scoring_array

    # TODO: Fix - This implementation results in the scoring array being generated every time we splice the sequence, which is not optimal.
    def zdna_calculator_transitions(self) -> np.ndarray:
        """
        Calculates the scoring array for a given sequence, treating each transition individually
        :return: np.ndarray: Scoring array
        """
        scoring_array = np.empty(len(self) - 1, dtype=float)
        mismatches_counter = 0  # mismatches_counter is the number of mismatches
        consecutive_AT_counter = 0
        for i in range(len(self) - 1):
            match self[i] + self[i + 1]:
                case "GC" | "CG":
                    scoring_array[i] = self.params.GC_weight
                    mismatches_counter = 0  # Reset the number of mismatches
                case "GT" | "TG":
                    scoring_array[i] = self.params.GT_weight
                    mismatches_counter = 0
                case "AC" | "CA":
                    scoring_array[i] = self.params.AC_weight
                    mismatches_counter = 0
                case "AT" | "TA":
                    adjusted_weight = self.params.AT_weight
                    # If we have less consecutive ATs than the provided AT scoring list, use the corresponding element
                    # Else use the last element of the list as penalty
                    if consecutive_AT_counter < len(self.params.consecutive_AT_scoring):
                        adjusted_weight += self.params.consecutive_AT_scoring[consecutive_AT_counter]
                    else:
                        adjusted_weight += self.params.consecutive_AT_scoring[-1]
                    scoring_array[i] = adjusted_weight
                    consecutive_AT_counter += 1
                    mismatches_counter = 0
                case _:  # If there was a mismatch
                    mismatches_counter += 1
                    if self.params.mismatch_penalty_type == "exponential":
                        # Limit the maximum value of the mismatch penalty to avoid exploding values
                        scoring_array[i] = -self.params.mismatch_penalty_starting_value ** mismatches_counter if mismatches_counter < 15 else -32000
                    elif self.params.mismatch_penalty_type == "linear":
                        scoring_array[i] = -self.params.mismatch_penalty_starting_value - self.params.mismatch_penalty_linear_delta * (mismatches_counter-1)
                    else:
                        raise ValueError(f"Mismatch penalty type not recognized. Valid options are {Params.mismatch_penalty_choices}")
        return scoring_array

    def zdna_calculator_coverage(self):
        """
        Adithya's proposal
        :return:
        """
        s = str(self)
        weights = {"gc": self.params.GC_weight, "gt": self.params.GT_weight, "ac": self.params.AC_weight, "at": self.params.AT_weight}
        scores = np.full(len(self), fill_value=-3, dtype=float)

        state = 0

        for i in range(len(s)):
            if i == 0:
                if s[i] in ['G', 'A']:
                    state = -1
                elif s[i] in ['C', 'T']:
                    state = 1

            elif state == -1:
                if s[i] in ['C', 'T']:
                    state = 2
                    sub_str = s[i - 1:i + 1]
                    match sub_str:
                        case "GC":
                            scores[i - 1:i + 1] = [weights["gc"]] * 2
                            last_weight = weights["gc"]
                        case "GT":
                            scores[i - 1:i + 1] = [weights["gt"]] * 2
                            last_weight = weights["gt"]
                        case "AC":
                            scores[i - 1:i + 1] = [weights["ac"]] * 2
                            last_weight = weights["ac"]
                        case "AT":
                            scores[i - 1:i + 1] = [weights["at"]] * 2
                            last_weight = weights["at"]

            elif state == 1:
                if s[i] in ['G', 'A']:
                    state = -2
                    sub_str = s[i - 1:i + 1]
                    match sub_str:
                        case "CG":
                            scores[i - 1:i + 1] = [weights["gc"]] * 2
                            last_weight = weights["gc"]
                        case "TG":
                            scores[i - 1:i + 1] = [weights["gt"]] * 2
                            last_weight = weights["gt"]
                        case "CA":
                            scores[i - 1:i + 1] = [weights["ac"]] * 2
                            last_weight = weights["ac"]
                        case "TA":
                            scores[i - 1:i + 1] = [weights["at"]] * 2
                            last_weight = weights["at"]

            elif state == -2:
                if s[i] in ['C', 'T']:
                    state = 2
                    sub_str = s[i - 1:i + 1]
                    match sub_str:
                        case "GC":
                            scores[i - 1] = max(scores[i - 1], weights["gc"])
                            scores[i] = weights["gc"]
                            last_weight = weights["gc"]
                        case "GT":
                            scores[i - 1] = max(scores[i - 1], weights["gt"])
                            scores[i] = weights["gt"]
                            last_weight = weights["gt"]
                        case "AC":
                            scores[i - 1] = max(scores[i - 1], weights["ac"])
                            scores[i] = weights["ac"]
                            last_weight = weights["ac"]
                        case "AT":
                            scores[i - 1] = max(scores[i - 1], weights["at"])
                            scores[i] = weights["at"]
                            last_weight = weights["at"]
                elif s[i] in ['G', 'A']:
                    state = -3

            elif state == 2:
                if s[i] in ['G', 'A']:
                    state = -2
                    sub_str = s[i - 1:i + 1]
                    match sub_str:
                        case "CG":
                            scores[i - 1] = max(scores[i - 1], weights["gc"])
                            scores[i] = weights["gc"]
                            last_weight = weights["gc"]
                        case "TG":
                            scores[i - 1] = max(scores[i - 1], weights["gt"])
                            scores[i] = weights["gt"]
                            last_weight = weights["gt"]
                        case "CA":
                            scores[i - 1] = max(scores[i - 1], weights["ac"])
                            scores[i] = weights["ac"]
                            last_weight = weights["ac"]
                        case "TA":
                            scores[i - 1] = max(scores[i - 1], weights["at"])
                            scores[i] = weights["at"]
                            last_weight = weights["at"]
                elif s[i] in ['C', 'T']:
                    state = 3

            elif state == -3:
                if s[i] in ['G', 'A']:
                    state = -1
                elif s[i] in ['C', 'T']:
                    sub_str = s[i - 1:i + 1]
                    state = 2
                    match sub_str:
                        case "GC":
                            if weights["gc"] > last_weight:
                                scores[i - 2] = -3
                                scores[i - 1:i + 1] = [weights["gc"]] * 2
                            else:
                                scores[i] = weights["gc"]
                            last_weight = weights["gc"]
                        case "GT":
                            if weights["gt"] > last_weight:
                                scores[i - 2] = -3
                                scores[i - 1:i + 1] = [weights["gt"]] * 2
                            else:
                                scores[i] = weights["gt"]
                            last_weight = weights["gt"]
                        case "AC":
                            if weights["ac"] > last_weight:
                                scores[i - 2] = -3
                                scores[i - 1:i + 1] = [weights["ac"]] * 2
                            else:
                                scores[i] = weights["ac"]
                            last_weight = weights["ac"]
                        case "AT":
                            if weights["at"] > last_weight:
                                scores[i - 2] = -3
                                scores[i - 1:i + 1] = [weights["at"]] * 2
                            else:
                                scores[i] = weights['at']
                            last_weight = weights["at"]

            elif state == 3:
                if s[i] in ['C', 'T']:
                    state = 1
                elif s[i] in ['G', 'A']:
                    sub_str = s[i - 1:i + 1]
                    state = -2
                    match sub_str:
                        case "CG":
                            if weights["gc"] > last_weight:
                                scores[i - 2] = -3
                                scores[i - 1:i + 1] = [weights["gc"]] * 2
                            else:
                                scores[i] = weights["gc"]
                            last_weight = weights["gc"]
                        case "TG":
                            if weights["gt"] > last_weight:
                                scores[i - 2] = -3
                                scores[i - 1:i + 1] = [weights["gt"]] * 2
                            else:
                                scores[i] = weights["gt"]
                            last_weight = weights["gt"]
                        case "CA":
                            if weights["ac"] > last_weight:
                                scores[i - 2] = -3
                                scores[i - 1:i + 1] = [weights["ac"]] * 2
                            else:
                                scores[i] = weights["ac"]
                            last_weight = weights["ac"]
                        case "TA":
                            if weights["at"] > last_weight:
                                scores[i - 2] = -3
                                scores[i - 1:i + 1] = [weights["at"]] * 2
                            else:
                                scores[i] = weights["at"]
                            last_weight = weights["at"]

        for i in range(1, len(scores)):
            if scores[i] == -3 and scores[i - 1] < 0:
                scores[i] = scores[i - 1] - 2

        return scores

    # TODO: Convert to generator
    def subarrays_above_threshold(self) -> list[tuple[int, int, int, str]]:
        # Calculate scoring array here instead of at initialization to avoid calculating it unecessarily
        match self.params.method:
            case "transitions":
                self.scoring_array = self.zdna_calculator_transitions()
            case "coverage":
                self.scoring_array = self.zdna_calculator_coverage()
            case "layered":
                self.scoring_array = self.zdna_calculator_layered()
            case _:  # If the method is not recognized
                raise ValueError(f"Method {self.params.method} not recognized. Valid options are: {Params.method_choices}")


        subarrays_above_threshold = []
        max_ending_here = self.scoring_array[0]
        start_idx = end_idx = 0
        current_max = 0  # Stores the maximum value of the currently considered candidate array
        candidate_array = None
        for i in range(1, len(self.scoring_array)):
            num = self.scoring_array[i]
            if num >= max_ending_here + num:
                start_idx = i
                end_idx = i + 1  # +1 because we are looking at the next element of the sequence to determine the weight
                max_ending_here = num
            else:
                max_ending_here += num
                end_idx = i + 1  # +1 because we are looking at the next element of the sequence to determine the weight
            if max_ending_here >= self.params.threshold and current_max < max_ending_here:
                candidate_array = (start_idx, end_idx, max_ending_here, str(self[start_idx: end_idx + 1]))
                current_max = max_ending_here

            # If we have a candidate array and the max_ending_here value dropped below 0, there is no way this element
            # will be party of a subarray above the threshold, so we add it to the list and reset the candidate array
            # Additionally, if the drop between the highest seen value and the current value is above the threshold
            # we also want to stop
            if candidate_array and (max_ending_here < 0 or current_max - max_ending_here >= self.params.threshold):
                subarrays_above_threshold.append(candidate_array)
                candidate_array = None
                max_ending_here = current_max = 0

        # If we have finished iterating without finding a negative value towards the end, we add the last candidate array
        if candidate_array:
            subarrays_above_threshold.append(candidate_array)

        return subarrays_above_threshold


# if __name__ == '__main__':
#     #test_string = "TTCGGCGAACTTCGGCGAGC"
#     #test_string = "TCGAGCGCCAGCGCCAAGCGCAGAGCGCAGGAGCGCAGCGCAGAGCGCCAG"
#     #test_string = "TCGCGCGATCGCGCGCGCGCGCGCCGCGATATATCG"
    
#     test_string = "AAAAACTAGCGGTCCG"

#     # test_string = "CGCGCGTATATATAGCCCCCCCCCCCCAGAGAGGGGAAGAGAGGGCCCCCAGAGTATATAGCGGAGAGC"
#     params = Params(method="layered", mismatch_penalty_type="linear", mismatch_penalty_starting_value=3, threshold=5)
#     seq = ZDNACalculatorSeq(test_string, params=params)

#     print([x for x in seq])
#     print(seq.scoring_array)
#     print(sum(seq.scoring_array))
#     print(seq.subarrays_above_threshold())

#     tests = False
#     if tests:
#         print("-------------------------")
#         params = Params(mismatch_penalty_type="linear", mismatch_penalty_starting_value=3, threshold=0)
#         seq = ZDNACalculatorSeq(test_string, params=params)
#         seq.scoring_array = [2, 10, 2, 2, -3, -2, 2, 2, 3]
#         for elem in seq.subarrays_above_threshold():
#             print(elem[0], elem[1], elem[2], seq.scoring_array[elem[0]: elem[1]])


